"""Tests for public API advanced wide form functions."""

from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Any, Optional

import pytest
from parametrization import Parametrization as P
from shapely import box

from overturemaestro._exceptions import (
    HierarchyDepthOutOfBoundsWarning,
    NegativeHierarchyDepthError,
)
from overturemaestro.advanced_functions import (
    convert_bounding_box_to_wide_form_geodataframe,
    convert_bounding_box_to_wide_form_geodataframe_for_all_types,
    convert_bounding_box_to_wide_form_geodataframe_for_multiple_types,
    convert_bounding_box_to_wide_form_parquet,
)
from overturemaestro.advanced_functions.wide_form import (
    _generate_result_file_path,
    get_all_possible_column_names,
    get_theme_type_classification,
)
from overturemaestro.data_downloader import PYARROW_FILTER
from overturemaestro.release_index import MINIMAL_SUPPORTED_RELEASE_VERSION
from tests.conftest import TEST_RELEASE_VERSION, bbox


@pytest.fixture(scope="session")  # type: ignore
def wide_form_working_directory() -> Path:
    """Working directory for tests."""
    return Path("files/wide_form")


@pytest.fixture(autouse=True, scope="session")  # type: ignore
def clear_working_directory(wide_form_working_directory: Path) -> None:
    """Clear working directory."""
    for file_path in wide_form_working_directory.glob("**/*.parquet"):
        file_path.unlink()


@pytest.mark.parametrize(
    "theme_type_pair",
    list(get_theme_type_classification(release=TEST_RELEASE_VERSION).keys()),
)  # type: ignore
def test_theme_type_pairs(
    test_release_version: str, wide_form_working_directory: Path, theme_type_pair: tuple[str, str]
) -> None:
    """Test if all theme type pairs are working."""
    convert_bounding_box_to_wide_form_parquet(
        theme=theme_type_pair[0],
        type=theme_type_pair[1],
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
    )


@pytest.mark.parametrize(
    "hierarchy_value",
    [None, 1],
)  # type: ignore
def test_all_theme_type_pairs(
    test_release_version: str, wide_form_working_directory: Path, hierarchy_value: Optional[int]
) -> None:
    """Test if downloading all theme type pairs at once is working."""
    gdf = convert_bounding_box_to_wide_form_geodataframe_for_all_types(
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        hierarchy_depth=hierarchy_value,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
    )
    prepared_theme_type_prefixes = tuple(
        "|".join(t) for t in get_theme_type_classification(release=test_release_version).keys()
    )
    feature_columns = [
        column_name for column_name in gdf.columns if column_name not in ("id", "geometry")
    ]
    assert all(
        column_name.startswith(prepared_theme_type_prefixes) for column_name in feature_columns
    )
    assert (gdf.dtypes.loc[feature_columns] == "bool").all()


@P.parameters("theme_type_pairs", "pyarrow_filters", "hierarchy_value", "expectation")  # type: ignore
@P.case(
    "Basic example without filters",
    [("base", "water"), ("base", "infrastructure")],
    None,
    None,
    does_not_raise(),
)  # type: ignore
@P.case(
    "Basic example with filters",
    [("base", "water"), ("base", "infrastructure")],
    [[("subtype", "==", "river")], [("class", "==", "bridge")]],
    None,
    does_not_raise(),
)  # type: ignore
@P.case(
    "Empty example with filters",
    [("base", "water"), ("base", "infrastructure")],
    [[("subtype", "==", "nonexistent")], [("class", "==", "nonexistent")]],
    None,
    does_not_raise(),
)  # type: ignore
@P.case(
    "Empty pyarrow filters list",
    [("base", "water"), ("base", "infrastructure")],
    [],
    None,
    pytest.raises(ValueError),
)  # type: ignore
@P.case(
    "Shorter pyarrow filters list",
    [("places", "place"), ("base", "infrastructure")],
    [[("confidence", ">", 0.95)]],
    None,
    pytest.raises(ValueError),
)  # type: ignore
@P.case(
    "Longer pyarrow filters list",
    [("places", "place"), ("base", "infrastructure")],
    [[("confidence", ">", 0.95)], [("class", "==", "bridge")], [("confidence", ">", 0.95)]],
    None,
    pytest.raises(ValueError),
)  # type: ignore
@P.case(
    "Basic example with hierarchy depth",
    [("base", "water"), ("base", "land_cover")],
    None,
    1,
    does_not_raise(),
)  # type: ignore
@P.case(
    "Example with wrong hierarchy depth",
    [("base", "water"), ("base", "land_cover")],
    None,
    2,
    pytest.warns(HierarchyDepthOutOfBoundsWarning),
)  # type: ignore
def test_multiple_theme_type_pairs(
    test_release_version: str,
    wide_form_working_directory: Path,
    theme_type_pairs: list[tuple[str, str]],
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]],
    hierarchy_value: Optional[int],
    expectation: Any,
) -> None:
    """Test if downloading multiple theme type pairs at once is working."""
    with expectation:
        gdf = convert_bounding_box_to_wide_form_geodataframe_for_multiple_types(
            theme_type_pairs=theme_type_pairs,
            bbox=bbox(),
            release=test_release_version,
            working_directory=wide_form_working_directory,
            hierarchy_depth=hierarchy_value,
            pyarrow_filters=pyarrow_filters,
            verbosity_mode="verbose",
            ignore_cache=False,
            include_all_possible_columns=False,
        )
        prepared_theme_type_prefixes = tuple("|".join(t) for t in theme_type_pairs)
        feature_columns = [
            column_name for column_name in gdf.columns if column_name not in ("id", "geometry")
        ]
        assert all(
            column_name.startswith(prepared_theme_type_prefixes) for column_name in feature_columns
        )
        assert (gdf.dtypes.loc[feature_columns] == "bool").all()


@P.parameters("hierarchy_value", "theme_type_pairs", "expectation")  # type: ignore
@P.case("Empty value", None, [("base", "water")], does_not_raise())  # type: ignore
@P.case("Zero", 0, [("base", "water")], does_not_raise())  # type: ignore
@P.case("Negative", -1, [("base", "water")], pytest.raises(NegativeHierarchyDepthError))  # type: ignore
@P.case("First value", 1, [("base", "water")], does_not_raise())  # type: ignore
@P.case("Second value", 2, [("base", "water")], does_not_raise())  # type: ignore
@P.case("Third value", 3, [("base", "water")], pytest.warns(HierarchyDepthOutOfBoundsWarning))  # type: ignore
@P.case("Single value as list", [0], [("base", "water")], does_not_raise())  # type: ignore
@P.case("Multiple values wrong length", [0, 0], [("base", "water")], pytest.raises(ValueError))  # type: ignore
@P.case(
    "Multiple values same length",
    [0, 0],
    [("base", "water"), ("base", "infrastructure")],
    does_not_raise(),
)  # type: ignore
@P.case(
    "Multiple values negative wrong length",
    [-1, -1],
    [("base", "water")],
    pytest.raises(ValueError),
)  # type: ignore
@P.case(
    "Multiple values negative same length",
    [-1, -1],
    [("base", "water"), ("base", "infrastructure")],
    pytest.raises(NegativeHierarchyDepthError),
)  # type: ignore
@P.case(
    "Multiple values same length with warning",
    [99, 99],
    [("base", "water"), ("base", "infrastructure")],
    pytest.warns(HierarchyDepthOutOfBoundsWarning),
)  # type: ignore
def test_hierarchy_values(
    test_release_version: str,
    wide_form_working_directory: Path,
    hierarchy_value: Optional[int],
    theme_type_pairs: list[tuple[str, str]],
    expectation: Any,
) -> None:
    """Test if hierarchy values are parsed correctly."""
    with expectation:
        gdf = convert_bounding_box_to_wide_form_geodataframe_for_multiple_types(
            theme_type_pairs=theme_type_pairs,
            bbox=bbox(),
            hierarchy_depth=hierarchy_value,
            release=test_release_version,
            working_directory=wide_form_working_directory,
            verbosity_mode="verbose",
            ignore_cache=False,
            include_all_possible_columns=False,
        )
        assert "geometry" in gdf.columns
        feature_columns = [
            column_name for column_name in gdf.columns if column_name not in ("id", "geometry")
        ]
        assert all("|" in column_name for column_name in feature_columns)
        assert (gdf.dtypes.loc[feature_columns] == "bool").all()


@pytest.mark.parametrize(
    "theme_type_pair",
    [("places", "place"), ("base", "water")],
)  # type: ignore
def test_include_all_possible_columns_parameter(
    test_release_version: str, wide_form_working_directory: Path, theme_type_pair: tuple[str, str]
) -> None:
    """Check if include_all_possible_columns works as intended."""
    theme_value, type_value = theme_type_pair
    pruned_dataset = convert_bounding_box_to_wide_form_geodataframe(
        theme=theme_value,
        type=type_value,
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
    ).drop(columns="geometry")

    full_dataset = convert_bounding_box_to_wide_form_geodataframe(
        theme=theme_value,
        type=type_value,
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=True,
    ).drop(columns="geometry")

    all_possible_columns = set(
        get_all_possible_column_names(
            theme=theme_value, type=type_value, release=test_release_version
        )
    )
    pruned_columns = set(pruned_dataset.columns)
    full_columns = set(full_dataset.columns)

    assert all_possible_columns == full_columns
    assert len(pruned_columns) < len(full_columns)

    columns_difference = full_columns.difference(pruned_columns)

    assert full_dataset.sum().loc[list(columns_difference)].sum() == 0


def test_empty_region(
    test_release_version: str,
    wide_form_working_directory: Path,
) -> None:
    """Test if regions without data are properly parsed."""
    gdf = convert_bounding_box_to_wide_form_geodataframe(
        "places",
        "place",
        (-10, -10, -9.9, -9.9),
        release=test_release_version,
        working_directory=wide_form_working_directory,
    ).drop(columns="geometry")

    all_possible_columns = set(
        get_all_possible_column_names(theme="places", type="place", release=test_release_version)
    )

    assert gdf.empty
    assert set(gdf.columns) == all_possible_columns


def test_confidence_parameter(
    test_release_version: str,
    wide_form_working_directory: Path,
) -> None:
    """Test if confidence parameter for places is working."""
    default_confidence = convert_bounding_box_to_wide_form_geodataframe(
        theme="places",
        type="place",
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
    )
    higher_confidence = convert_bounding_box_to_wide_form_geodataframe(
        theme="places",
        type="place",
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
        places_minimal_confidence=0.95,
    )
    lower_confidence = convert_bounding_box_to_wide_form_geodataframe(
        theme="places",
        type="place",
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
        places_minimal_confidence=0.05,
    )

    assert len(default_confidence) > len(higher_confidence)
    assert len(default_confidence) < len(lower_confidence)


def test_places_use_primary_category_only_parameter(
    test_release_version: str,
    wide_form_working_directory: Path,
) -> None:
    """Test if places_use_primary_category_only parameter is working."""
    primary_only = convert_bounding_box_to_wide_form_geodataframe(
        theme="places",
        type="place",
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
        places_use_primary_category_only=True,
    ).drop(columns="geometry")

    all_categories = convert_bounding_box_to_wide_form_geodataframe(
        theme="places",
        type="place",
        bbox=bbox(),
        release=test_release_version,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
        places_use_primary_category_only=False,
    ).drop(columns="geometry")

    assert (primary_only.sum(axis=1) == 1).all(), "Not all rows have exactly one primary category."
    assert (
        primary_only.sum().sum() < all_categories.sum().sum()
    ), "Primary only has more categories than all categories."


def test_old_version(
    wide_form_working_directory: Path,
) -> None:
    """Test if oldest supported version is working."""
    convert_bounding_box_to_wide_form_geodataframe_for_all_types(
        bbox=bbox(),
        release=MINIMAL_SUPPORTED_RELEASE_VERSION,
        working_directory=wide_form_working_directory,
        verbosity_mode="verbose",
        ignore_cache=False,
        include_all_possible_columns=False,
    )


def test_generate_result_file_name_order(
    test_release_version: str,
) -> None:
    """Test if result file name is generated correctly."""
    theme_type_pairs = [
        ("base", "water"),
        ("base", "land_cover"),
        ("base", "infrastructure"),
        ("places", "place"),
    ]
    pyarrow_filters = [None, None, [("class", "==", "bridge")], [("confidence", ">", 0.95)]]
    hierarchy_depths = [None, 1, 2, 3]
    result = _generate_result_file_path(
        release=test_release_version,
        theme_type_pairs=theme_type_pairs,
        geometry_filter=box(*bbox()),
        pyarrow_filters=pyarrow_filters,
        hierarchy_depth=hierarchy_depths,
        include_all_possible_columns=False,
    )

    reverse_order_result = _generate_result_file_path(
        release=test_release_version,
        theme_type_pairs=theme_type_pairs[::-1],
        geometry_filter=box(*bbox()),
        pyarrow_filters=pyarrow_filters[::-1],
        hierarchy_depth=hierarchy_depths[::-1],
        include_all_possible_columns=False,
    )

    assert result == reverse_order_result



"""Functions for retrieving Overture Maps features in a wide form."""

import json
import tempfile
import warnings
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Protocol, Union, overload

import numpy as np
import pandas as pd
from fsspec.implementations.http import HTTPFileSystem
from pooch import file_hash, retrieve
from pooch import get_logger as get_pooch_logger
from requests import HTTPError
from rich import print as rprint
from shapely.geometry.base import BaseGeometry

from overturemaestro._duckdb import _set_up_duckdb_connection, _sql_escape
from overturemaestro._exceptions import (
    HierarchyDepthOutOfBoundsWarning,
    NegativeHierarchyDepthError,
)
from overturemaestro._rich_progress import VERBOSITY_MODE, TrackProgressBar, TrackProgressSpinner
from overturemaestro.cache import (
    _get_global_wide_form_release_cache_directory,
    _get_local_wide_form_release_cache_directory,
)
from overturemaestro.data_downloader import (
    PYARROW_FILTER,
    _generate_geometry_hash,
    download_data_for_multiple_types,
)
from overturemaestro.elapsed_time_decorator import show_total_elapsed_time_decorator
from overturemaestro.release_index import (
    LFS_DIRECTORY_URL,
    _check_release_version,
    get_newest_release_version,
)

__all__ = [
    "convert_geometry_to_wide_form_parquet_for_all_types",
    "convert_geometry_to_wide_form_parquet_for_multiple_types",
    "get_all_possible_column_names",
]

if TYPE_CHECKING:  # pragma: no cover
    from pandas import DataFrame
    from pyarrow.compute import Expression


def _check_depth_for_wide_form(
    theme: str, type: str, hierarchy_columns: list[str], depth: Optional[int] = None, **kwargs: Any
) -> int:
    depth = depth if depth is not None else len(hierarchy_columns)

    if depth < 0:
        raise NegativeHierarchyDepthError("Hierarchy depth cannot be negative")
    elif depth > len(hierarchy_columns):
        warnings.warn(
            (
                f"Provided hierarchy depth is out of bounds"
                f" (valid for {theme}/{type}: 0 - {len(hierarchy_columns)})."
                f" Value will be clipped to {len(hierarchy_columns)}."
            ),
            HierarchyDepthOutOfBoundsWarning,
            stacklevel=0,
        )
        depth = len(hierarchy_columns)

    return depth


def _transform_to_wide_form(
    theme: str,
    type: str,
    release_version: str,
    parquet_file: Path,
    output_path: Path,
    include_all_possible_columns: bool,
    hierarchy_columns: list[str],
    working_directory: Union[str, Path],
    verbosity_mode: VERBOSITY_MODE,
    **kwargs: Any,
) -> Path:
    connection = _set_up_duckdb_connection(working_directory)

    if include_all_possible_columns:
        wide_column_definitions = _get_wide_column_definitions(
            theme=theme,
            type=type,
            release_version=release_version,
            hierarchy_columns=hierarchy_columns,
            verbosity_mode="silent",
        ).to_dict(orient="records")
    else:
        joined_hierarchy_columns = ",".join(hierarchy_columns)

        available_colums_sql_query = f"""
        SELECT DISTINCT
            {joined_hierarchy_columns},
            concat_ws('|', '{theme}', '{type}', {joined_hierarchy_columns}) as column_name
        FROM read_parquet(
            '{parquet_file}',
            hive_partitioning=false
        )
        ORDER BY column_name
        """

        wide_column_definitions = (
            connection.sql(available_colums_sql_query).fetchdf().to_dict(orient="records")
        )

    case_clauses = []

    for wide_column_definition in wide_column_definitions:
        column_name = wide_column_definition["column_name"] or f"{theme}|{type}"
        conditions = []
        for condition_column in hierarchy_columns:
            if wide_column_definition[condition_column] is None:
                conditions.append(f"{condition_column} IS NULL")
            else:
                escaped_value = _sql_escape(wide_column_definition[condition_column])
                conditions.append(f"{condition_column} = '{escaped_value}'")

        joined_conditions = " AND ".join(conditions)
        case_clauses.append(f'COALESCE(({joined_conditions}), False) AS "{column_name}"')

    query = f"""
    COPY (
        SELECT
            id,
            geometry,
            {", ".join(case_clauses)}
        FROM read_parquet(
            '{parquet_file}',
            hive_partitioning=false
        )
    ) TO '{output_path}' (
        FORMAT 'parquet',
        PER_THREAD_OUTPUT false
    )
    """

    connection.execute(query)

    return output_path


def _transform_to_wide_form_without_hierarchy(
    theme: str,
    type: str,
    parquet_file: Path,
    output_path: Path,
    working_directory: Union[str, Path],
    **kwargs: Any,
) -> Path:
    connection = _set_up_duckdb_connection(working_directory)

    query = f"""
    COPY (
        SELECT
            id,
            geometry,
            true as "{theme}|{type}"
        FROM read_parquet(
            '{parquet_file}',
            hive_partitioning=false
        )
    ) TO '{output_path}' (
        FORMAT 'parquet',
        PER_THREAD_OUTPUT false
    )
    """

    connection.execute(query)

    return output_path


def _prepare_download_parameters_for_poi(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    hierachy_columns: list[str],
    pyarrow_filter: Optional["Expression"] = None,
    **kwargs: Any,
) -> tuple[list[str], Optional["Expression"]]:
    # TODO: swap to dedicated function?
    import pyarrow.compute as pc

    category_not_null_filter = pc.invert(pc.field("categories").is_null())
    minimal_confidence_filter = pc.field("confidence") >= pc.scalar(
        kwargs.get("places_minimal_confidence", 0.75)
    )
    if pyarrow_filter is not None:
        pyarrow_filter = pyarrow_filter & category_not_null_filter & minimal_confidence_filter
    else:
        pyarrow_filter = category_not_null_filter & minimal_confidence_filter

    return (["categories"] if hierachy_columns else [], pyarrow_filter)


def _transform_poi_to_wide_form(
    theme: str,
    type: str,
    release_version: str,
    parquet_file: Path,
    output_path: Path,
    include_all_possible_columns: bool,
    hierarchy_columns: list[str],
    working_directory: Union[str, Path],
    verbosity_mode: VERBOSITY_MODE,
    **kwargs: Any,
) -> Path:
    connection = _set_up_duckdb_connection(working_directory)

    primary_category_only = kwargs.get("places_use_primary_category_only", False)

    primary_category_name = "primary"
    alternate_category_name = "alternate"
    if release_version < "2024-07-22.0":
        primary_category_name = "main"

    wide_column_definitions = _get_wide_column_definitions_for_poi(
        theme=theme,
        type=type,
        release_version=release_version,
        hierarchy_columns=hierarchy_columns,
        verbosity_mode="silent",
    )

    if not include_all_possible_columns:
        if primary_category_only:
            available_colums_sql_query = f"""
            SELECT DISTINCT
                categories.{primary_category_name} as category
            FROM read_parquet(
                '{parquet_file}',
                hive_partitioning=false
            )
            """
        else:
            available_colums_sql_query = f"""
            SELECT DISTINCT
                categories.{primary_category_name} as category
            FROM read_parquet(
                '{parquet_file}',
                hive_partitioning=false
            )
            UNION
            SELECT DISTINCT
                UNNEST(categories.{alternate_category_name}) as category
            FROM read_parquet(
                '{parquet_file}',
                hive_partitioning=false
            )
            """

        existing_category_names = (
            connection.sql(available_colums_sql_query).fetchdf()["category"].sort_values()
        )
        wide_column_definitions = wide_column_definitions[
            wide_column_definitions["category"].isin(existing_category_names)
        ]

    case_clauses = []

    for column_name, categories_list in (
        wide_column_definitions.groupby("column_name")["category"].agg(list).to_dict().items()
    ):
        conditions = []
        for category_name in categories_list:
            escaped_value = _sql_escape(category_name)
            conditions.append(f"categories.{primary_category_name} = '{escaped_value}'")

            if not primary_category_only:
                conditions.append(f"'{escaped_value}' IN categories.{alternate_category_name}")

        joined_conditions = " OR ".join(conditions)
        case_clauses.append(f'COALESCE(({joined_conditions}), False) AS "{column_name}"')

    query = f"""
    COPY (
        SELECT
            id,
            geometry,
            {", ".join(case_clauses)}
        FROM read_parquet(
            '{parquet_file}',
            hive_partitioning=false
        )
    ) TO '{output_path}' (
        FORMAT 'parquet',
        PER_THREAD_OUTPUT false
    )
    """

    connection.execute(query)

    return output_path


def _get_all_possible_column_names(
    theme: str, type: str, release_version: str, hierarchy_columns: list[str], **kwargs: Any
) -> "DataFrame":
    import duckdb

    connection = duckdb.connect()

    for extension in ("spatial", "httpfs"):
        connection.install_extension(extension)
        connection.load_extension(extension)

    dataset_path = (
        f"s3://overturemaps-us-west-2/release/{release_version}/theme={theme}/type={type}/*"
    )
    joined_hierarchy_columns = ",".join(hierarchy_columns)
    df = duckdb.sql(
        f"""
        SELECT DISTINCT
            {joined_hierarchy_columns}
        FROM read_parquet(
            '{dataset_path}',
            hive_partitioning=false
        )
        ORDER BY ALL
        """
    ).to_df()

    return df


def _get_all_possible_column_names_for_poi(
    theme: str, type: str, release_version: str, hierarchy_columns: list[str], **kwargs: Any
) -> "DataFrame":
    import duckdb

    connection = duckdb.connect()

    for extension in ("spatial", "httpfs"):
        connection.install_extension(extension)
        connection.load_extension(extension)

    dataset_path = (
        f"s3://overturemaps-us-west-2/release/{release_version}/theme={theme}/type={type}/*"
    )

    primary_category_name = "primary"
    alternate_category_name = "alternate"
    if release_version < "2024-07-22.0":
        primary_category_name = "main"

    df = (
        duckdb.sql(
            f"""
        SELECT DISTINCT
            categories.{primary_category_name} as column_name
        FROM read_parquet(
            '{dataset_path}',
            hive_partitioning=false
        )
        UNION
        SELECT DISTINCT
            UNNEST(categories.{alternate_category_name}) as column_name
        FROM read_parquet(
            '{dataset_path}',
            hive_partitioning=false
        )
        """
        )
        .to_df()
        .dropna()
        .sort_values(by="column_name")
        .reset_index(drop=True)
    )

    hierarchy_data = pd.read_csv(
        # List of all possible places values on CC-BY-SA 4.0 license
        # provided by Overture Maps Foundation
        "https://raw.githubusercontent.com/OvertureMaps/schema/refs/heads/main/docs/schema/concepts/by-theme/places/overture_categories.csv",
        sep=";",
        names=["category", "hierarchy"],
        skiprows=1,
    )
    hierarchy_split = (
        hierarchy_data["hierarchy"].str.strip().str[1:-1].str.split(",").apply(pd.Series)
    )

    hierarchy_split.columns = [str(i + 1) for i in range(hierarchy_split.shape[1])]

    # Concatenate the original dataframe with the new hierarchy columns
    data_split = pd.concat([hierarchy_data[["category"]], hierarchy_split], axis=1)

    rows = data_split.to_dict(orient="records")

    for category_name in df["column_name"]:
        if category_name not in data_split["category"].values:
            rows.append({"category": category_name, "1": category_name})

    return pd.DataFrame(rows).replace({np.nan: None})


def _get_wide_column_definitions(
    theme: str,
    type: str,
    release_version: str,
    hierarchy_columns: list[str],
    verbosity_mode: VERBOSITY_MODE = "transient",
    **kwargs: Any,
) -> "DataFrame":
    if not hierarchy_columns:
        return pd.DataFrame(dict(column_name=[f"{theme}|{type}"]))

    all_columns_names = load_wide_form_all_column_names_release_index(
        theme=theme, type=type, release=release_version, verbosity_mode=verbosity_mode
    )
    columns_not_in_hierarchy = [c for c in all_columns_names.columns if c not in hierarchy_columns]
    if columns_not_in_hierarchy:
        all_columns_names = all_columns_names.drop(
            columns=columns_not_in_hierarchy
        ).drop_duplicates()
    combine_columns_fn = partial(
        _combine_columns, theme=theme, type=type, hierarchy_columns=hierarchy_columns
    )
    all_columns_names["column_name"] = all_columns_names.apply(combine_columns_fn, axis=1)
    return all_columns_names.sort_values(by="column_name")


def _combine_columns(row: pd.Series, theme: str, type: str, hierarchy_columns: list[str]) -> str:
    result = f"{theme}|{type}"
    for column_name in hierarchy_columns:
        value = row[column_name]
        if value is None:
            break
        result += f"|{value}"
    return result


def _get_wide_column_definitions_for_poi(
    theme: str,
    type: str,
    release_version: str,
    hierarchy_columns: list[str],
    verbosity_mode: VERBOSITY_MODE = "transient",
    **kwargs: Any,
) -> "DataFrame":
    if not hierarchy_columns:
        return pd.DataFrame(dict(column_name=[f"{theme}|{type}"]))

    all_columns_names = load_wide_form_all_column_names_release_index(
        theme=theme, type=type, release=release_version, verbosity_mode=verbosity_mode
    )
    columns_not_in_hierarchy = [
        c for c in all_columns_names.columns if c != "category" and c not in hierarchy_columns
    ]
    if columns_not_in_hierarchy:
        all_columns_names = all_columns_names.drop(
            columns=columns_not_in_hierarchy
        ).drop_duplicates()
    combine_columns_fn = partial(
        _combine_columns, theme=theme, type=type, hierarchy_columns=hierarchy_columns
    )
    all_columns_names["column_name"] = all_columns_names.apply(combine_columns_fn, axis=1)
    return all_columns_names.sort_values(by="column_name")


class DownloadParametersPreparationCallable(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self,
        theme: str,
        type: str,
        geometry_filter: BaseGeometry,
        hierachy_columns: list[str],
        pyarrow_filter: Optional["Expression"],
        **kwargs: Any,
    ) -> tuple[list[str], Optional["Expression"]]: ...


class DepthCheckCallable(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self,
        theme: str,
        type: str,
        hierarchy_columns: list[str],
        depth: Optional[int],
        **kwargs: Any,
    ) -> int: ...


class DataTransformationCallable(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self,
        theme: str,
        type: str,
        release_version: str,
        parquet_file: Path,
        output_path: Path,
        include_all_possible_columns: bool,
        hierarchy_columns: list[str],
        working_directory: Union[str, Path],
        verbosity_mode: VERBOSITY_MODE,
        **kwargs: Any,
    ) -> Path: ...


class GetAllPossibleColumnNamesCallable(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self,
        theme: str,
        type: str,
        release_version: str,
        hierarchy_columns: list[str],
        **kwargs: Any,
    ) -> "DataFrame": ...


class GetWideColumnDefinitionsCallable(Protocol):  # noqa: D101
    def __call__(  # noqa: D102
        self,
        theme: str,
        type: str,
        release_version: str,
        hierarchy_columns: list[str],
        verbosity_mode: VERBOSITY_MODE,
        **kwargs: Any,
    ) -> "DataFrame": ...


@dataclass
class WideFormDefinition:
    """Helper object for theme type wide form logic definition."""

    hierachy_columns: list[str]
    download_parameters_preparation_function: Optional[DownloadParametersPreparationCallable] = None
    depth_check_function: DepthCheckCallable = _check_depth_for_wide_form
    data_transform_function: DataTransformationCallable = _transform_to_wide_form
    get_all_possible_column_names_function: GetAllPossibleColumnNamesCallable = (
        _get_all_possible_column_names
    )
    get_wide_column_definitions_function: GetWideColumnDefinitionsCallable = (
        _get_wide_column_definitions
    )


THEME_TYPE_CLASSIFICATION: dict[tuple[str, str], WideFormDefinition] = {
    ("base", "infrastructure"): WideFormDefinition(hierachy_columns=["subtype", "class"]),
    ("base", "land"): WideFormDefinition(hierachy_columns=["subtype", "class"]),
    ("base", "land_cover"): WideFormDefinition(hierachy_columns=["subtype"]),
    ("base", "land_use"): WideFormDefinition(hierachy_columns=["subtype", "class"]),
    ("base", "water"): WideFormDefinition(hierachy_columns=["subtype", "class"]),
    ("transportation", "segment"): WideFormDefinition(
        hierachy_columns=["subtype", "class", "subclass"]
    ),
    ("places", "place"): WideFormDefinition(
        hierachy_columns=["1", "2", "3", "4", "5", "6"],  # 6 IS A MAX DEPTH
        download_parameters_preparation_function=_prepare_download_parameters_for_poi,
        data_transform_function=_transform_poi_to_wide_form,
        get_all_possible_column_names_function=_get_all_possible_column_names_for_poi,
        get_wide_column_definitions_function=_get_wide_column_definitions_for_poi,
    ),
    ("buildings", "building"): WideFormDefinition(hierachy_columns=["subtype", "class"]),
}


def get_theme_type_classification(release: str) -> dict[tuple[str, str], WideFormDefinition]:
    classification = THEME_TYPE_CLASSIFICATION.copy()
    # start from the newest release

    if release < "2024-08-20.0":
        classification[("transportation", "segment")] = WideFormDefinition(
            hierachy_columns=["subtype", "class"]
        )

    if release < "2024-05-16-beta.0":
        classification[("buildings", "building")] = WideFormDefinition(hierachy_columns=["class"])
        classification.pop(("base", "land_cover"))

    return classification


@overload
def convert_geometry_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_geometry_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_geometry_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@show_total_elapsed_time_decorator
def convert_geometry_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path:
    """
    Get GeoParquet file for a given geometry in a wide format for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a single file as a result with multiple columns based
    on dataset schema.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only primary category
            from the places dataset. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        Path: Path to the generated GeoParquet file.
    """
    if pyarrow_filters is not None and len(theme_type_pairs) != len(pyarrow_filters):
        raise ValueError("Pyarrow filters length doesn't match length of theme type pairs.")

    if isinstance(hierarchy_depth, list) and len(theme_type_pairs) != len(hierarchy_depth):
        raise ValueError("Hierarchy depth list length doesn't match length of theme type pairs.")

    if not release:
        release = get_newest_release_version()

    pyarrow_filters_list = []
    for idx in range(len(theme_type_pairs)):
        _pyarrow_filter = pyarrow_filters[idx] if pyarrow_filters else None

        if _pyarrow_filter is not None:
            from pyarrow.parquet import filters_to_expression

            _pyarrow_filter = filters_to_expression(_pyarrow_filter)

        pyarrow_filters_list.append(_pyarrow_filter)

    if result_file_path is None:
        result_file_path = working_directory / _generate_result_file_path(
            release=release,
            theme_type_pairs=theme_type_pairs,
            geometry_filter=geometry_filter,
            include_all_possible_columns=include_all_possible_columns,
            hierarchy_depth=hierarchy_depth,
            pyarrow_filters=pyarrow_filters_list,
            places_use_primary_category_only=places_use_primary_category_only,
            places_minimal_confidence=places_minimal_confidence,
        )

    result_file_path = Path(result_file_path)

    if not result_file_path.exists() or ignore_cache:
        result_file_path.parent.mkdir(exist_ok=True, parents=True)

        prepared_download_parameters = _prepare_download_parameters_for_all_theme_type_pairs(
            release=release,
            theme_type_pairs=theme_type_pairs,
            geometry_filter=geometry_filter,
            hierarchy_depth=hierarchy_depth,
            pyarrow_filters=pyarrow_filters_list,
            verbosity_mode=verbosity_mode,
            places_minimal_confidence=places_minimal_confidence,
        )

        hierachy_columns_list, columns_to_download_list, pyarrow_filter_list = zip(
            *prepared_download_parameters
        )

        downloaded_parquet_files = download_data_for_multiple_types(
            release=release,
            theme_type_pairs=theme_type_pairs,
            geometry_filter=geometry_filter,
            pyarrow_filters=pyarrow_filter_list,
            columns_to_download=[
                ["id", "geometry", *columns_to_download]
                for columns_to_download in columns_to_download_list
            ],
            ignore_cache=ignore_cache,
            working_directory=working_directory,
            verbosity_mode=verbosity_mode,
            max_workers=max_workers,
        )

        with tempfile.TemporaryDirectory(dir=Path(working_directory).resolve()) as tmp_dir_name:
            tmp_dir_path = Path(tmp_dir_name)

            transformed_wide_form_directory_output = tmp_dir_path / "wide_form_files"
            transformed_wide_form_directory_output.mkdir(parents=True, exist_ok=True)

            with TrackProgressBar(verbosity_mode=verbosity_mode) as progress:
                for (
                    (theme_value, type_value),
                    hierachy_columns,
                    downloaded_parquet_file,
                ) in progress.track(
                    zip(theme_type_pairs, hierachy_columns_list, downloaded_parquet_files),
                    total=len(theme_type_pairs),
                    description="Transforming data into wide form",
                ):
                    wide_form_definition = get_theme_type_classification(release=release)[
                        (theme_value, type_value)
                    ]

                    output_path = (
                        transformed_wide_form_directory_output
                        / f"{theme_value}_{type_value}.parquet"
                    )
                    if len(theme_type_pairs) == 1:
                        output_path = result_file_path

                    if not hierachy_columns:
                        _transform_to_wide_form_without_hierarchy(
                            theme=theme_value,
                            type=type_value,
                            parquet_file=downloaded_parquet_file,
                            output_path=output_path,
                            working_directory=tmp_dir_path,
                        )
                    else:
                        wide_form_definition.data_transform_function(
                            theme=theme_value,
                            type=type_value,
                            release_version=release,
                            parquet_file=downloaded_parquet_file,
                            output_path=output_path,
                            include_all_possible_columns=include_all_possible_columns,
                            hierarchy_columns=hierachy_columns,
                            working_directory=tmp_dir_path,
                            verbosity_mode=verbosity_mode,
                            places_use_primary_category_only=places_use_primary_category_only,
                        )

            if len(theme_type_pairs) > 1:
                with TrackProgressSpinner(
                    "Joining results to a single file", verbosity_mode=verbosity_mode
                ):
                    _combine_multiple_wide_form_files(
                        theme_type_pairs=theme_type_pairs,
                        transformed_wide_form_directory=transformed_wide_form_directory_output,
                        output_path=result_file_path,
                        working_directory=tmp_dir_path,
                    )

    return result_file_path


@overload
def convert_geometry_to_wide_form_parquet_for_all_types(
    geometry_filter: BaseGeometry,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_geometry_to_wide_form_parquet_for_all_types(
    geometry_filter: BaseGeometry,
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_geometry_to_wide_form_parquet_for_all_types(
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


def convert_geometry_to_wide_form_parquet_for_all_types(
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path:
    """
    Get GeoParquet file for a given geometry in a wide format for all types.

    Automatically downloads Overture Maps dataset for a given release and all available theme/types
    in a concurrent manner and returns a single file as a result with multiple columns based
    on dataset schema.

    Args:
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only primary category
            from the places dataset. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        Path: Path to the generated GeoParquet file.
    """
    if not release:
        release = get_newest_release_version()

    return convert_geometry_to_wide_form_parquet_for_multiple_types(
        theme_type_pairs=list(get_theme_type_classification(release=release).keys()),
        geometry_filter=geometry_filter,
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


def get_all_possible_column_names(
    release: Optional[str] = None,
    theme: Optional[str] = None,
    type: Optional[str] = None,
    hierarchy_depth: Optional[int] = None,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> list[str]:
    """
    Get a list of all possible columns.

    Args:
        release (Optional[str], optional): Select the release for which the list should be returned.
            If None, will return for the newest release. Defaults to None.
        theme (Optional[str], optional): Select the theme of the dataset. If None, will return
            the list for all datasets. Defaults to None.
        type (Optional[str], optional): Select the type of the dataset. If None, will return
            the list for all datasets. Defaults to None.
        hierarchy_depth (Optional[int]): Depth used to calculate how many hierarchy columns should
            be used to generate the wide form of the data. If None, will use all available columns.
            Defaults to None.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".

    Returns:
        list[str]: List of column names.
    """
    if (theme is None and type is not None) or (theme is not None and type is None):
        raise ValueError("Theme and type both have to be present or None.")

    if not release:
        release = get_newest_release_version()

    _check_release_version(release)

    if theme and type:
        definitions = {(theme, type): get_theme_type_classification(release=release)[(theme, type)]}
    else:
        definitions = get_theme_type_classification(release=release)

    columns = []
    for (theme_value, type_value), wide_form_definition in definitions.items():
        depth = wide_form_definition.depth_check_function(
            theme=theme_value,
            type=type_value,
            hierarchy_columns=wide_form_definition.hierachy_columns,
            depth=hierarchy_depth,
        )
        hierachy_columns = wide_form_definition.hierachy_columns[:depth]

        df = wide_form_definition.get_wide_column_definitions_function(
            theme=theme_value,
            type=type_value,
            release_version=release,
            hierarchy_columns=hierachy_columns,
            verbosity_mode=verbosity_mode,
        )
        columns.extend(df["column_name"].unique())

    return sorted(columns)


def _generate_result_file_path(
    release: str,
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    include_all_possible_columns: bool,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]],
    pyarrow_filters: Optional[list[Union["Expression", None]]],
    **kwargs: Any,
) -> Path:
    import hashlib

    directory = Path(release)

    params_per_theme_type_pair = {
        theme_type_pair: (
            hierarchy_depth[idx] if isinstance(hierarchy_depth, list) else hierarchy_depth,
            pyarrow_filters[idx] if pyarrow_filters else None,
        )
        for idx, theme_type_pair in enumerate(theme_type_pairs)
    }

    sorted_theme_type_pairs = sorted(theme_type_pairs)

    if len(theme_type_pairs) == 1:
        theme_value, type_value = theme_type_pairs[0]
        directory = directory / f"theme={theme_value}" / f"type={type_value}"
    else:
        h = hashlib.new("sha256")
        h.update(str(sorted_theme_type_pairs).encode())
        directory = directory / h.hexdigest()

    clipping_geometry_hash_part = _generate_geometry_hash(geometry_filter)

    pyarrow_filter_hash_part = "nofilter"
    if pyarrow_filters is not None:
        h = hashlib.new("sha256")
        for theme_type_pair in sorted_theme_type_pairs:
            h.update(str(params_per_theme_type_pair[theme_type_pair][1]).encode())
        pyarrow_filter_hash_part = h.hexdigest()[:8]

    hierarchy_hash_part = ""
    if hierarchy_depth is not None:
        h = hashlib.new("sha256")
        for theme_type_pair in sorted_theme_type_pairs:
            h.update(str(params_per_theme_type_pair[theme_type_pair][0]).encode())
        hierarchy_hash_part = f"_h{h.hexdigest()[:8]}"

    include_all_columns_hash_part = ""
    if not include_all_possible_columns:
        include_all_columns_hash_part = "_pruned"

    h = hashlib.new("sha256")
    h.update(str(sorted(kwargs.items())).encode())
    kwargs_hash_part = h.hexdigest()[:8]

    return directory / (
        f"{clipping_geometry_hash_part}_{pyarrow_filter_hash_part}_{kwargs_hash_part}"
        f"_wide_form{hierarchy_hash_part}{include_all_columns_hash_part}.parquet"
    )


def _prepare_download_parameters_for_all_theme_type_pairs(
    release: str,
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Union["Expression", None]]] = None,
    verbosity_mode: VERBOSITY_MODE = "transient",
    **kwargs: Any,
) -> list[tuple[list[str], list[str], Optional[PYARROW_FILTER]]]:
    prepared_parameters = []
    with TrackProgressBar(verbosity_mode=verbosity_mode) as progress:
        for idx, (theme_value, type_value) in progress.track(
            enumerate(theme_type_pairs),
            total=len(theme_type_pairs),
            description="Preparing download parameters",
        ):
            single_pyarrow_filter = pyarrow_filters[idx] if pyarrow_filters else None
            single_hierarchy_depth = (
                hierarchy_depth[idx] if isinstance(hierarchy_depth, list) else hierarchy_depth
            )
            wide_form_definition = get_theme_type_classification(release=release)[
                (theme_value, type_value)
            ]

            depth = wide_form_definition.depth_check_function(
                theme=theme_value,
                type=type_value,
                hierarchy_columns=wide_form_definition.hierachy_columns,
                depth=single_hierarchy_depth,
                **kwargs,
            )
            hierachy_columns = wide_form_definition.hierachy_columns[:depth]
            columns_to_download = hierachy_columns

            if wide_form_definition.download_parameters_preparation_function is not None:
                columns_to_download, single_pyarrow_filter = (
                    wide_form_definition.download_parameters_preparation_function(
                        theme=theme_value,
                        type=type_value,
                        geometry_filter=geometry_filter,
                        hierachy_columns=hierachy_columns,
                        pyarrow_filter=single_pyarrow_filter,
                        **kwargs,
                    )
                )

            prepared_parameters.append(
                (hierachy_columns, columns_to_download, single_pyarrow_filter)
            )

    return prepared_parameters


def _combine_multiple_wide_form_files(
    theme_type_pairs: list[tuple[str, str]],
    transformed_wide_form_directory: Path,
    output_path: Path,
    working_directory: Union[str, Path],
) -> None:
    import pyarrow.parquet as pq

    all_select_columns = []
    subqueries = []

    for theme_value, type_value in theme_type_pairs:
        current_parquet_file = (
            transformed_wide_form_directory / f"{theme_value}_{type_value}.parquet"
        )
        available_columns = [
            col
            for col in pq.read_metadata(current_parquet_file).schema.names
            if col not in ("id", "geometry")
        ]

        combined_columns_select = ", ".join(f'"{column}"' for column in available_columns)

        select_subquery = f"""
        SELECT id, geometry, {combined_columns_select}
        FROM read_parquet('{current_parquet_file}', hive_partitioning=false)
        """

        subqueries.append(select_subquery)
        all_select_columns.extend(available_columns)

    joined_all_select_columns = ", ".join(
        f'COALESCE("{col}", False) AS "{col}"' for col in sorted(all_select_columns)
    )
    joined_subqueries = " UNION ALL BY NAME ".join(subqueries)

    connection = _set_up_duckdb_connection(working_directory)
    query = f"""
    COPY (
        SELECT
            id,
            geometry,
            {joined_all_select_columns}
        FROM (
            {joined_subqueries}
        )
    ) TO '{output_path}' (
        FORMAT 'parquet',
        PER_THREAD_OUTPUT false
    )
    """

    connection.execute(query)


@overload
def load_wide_form_all_column_names_release_index(
    theme: str,
    type: str,
    *,
    remote_index: bool = False,
    skip_index_download: bool = False,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> "DataFrame": ...


@overload
def load_wide_form_all_column_names_release_index(
    theme: str,
    type: str,
    release: str,
    *,
    remote_index: bool = False,
    skip_index_download: bool = False,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> "DataFrame": ...


@overload
def load_wide_form_all_column_names_release_index(
    theme: str,
    type: str,
    release: Optional[str] = None,
    *,
    remote_index: bool = False,
    skip_index_download: bool = False,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> "DataFrame": ...


def load_wide_form_all_column_names_release_index(
    theme: str,
    type: str,
    release: Optional[str] = None,
    *,
    remote_index: bool = False,
    skip_index_download: bool = False,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> "DataFrame":
    """
    Load release index as a GeoDataFrame.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        geometry_filter (Optional[BaseGeometry], optional): Geometry to pre-filter resulting rows.
            Defaults to None.
        remote_index (bool, optional): Avoid downloading the index and stream it from remote source.
            Defaults to False.
        skip_index_download (bool, optional): Avoid downloading the index if doesn't exist locally
            and generate it instead. Defaults to False.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".

    Returns:
        gpd.GeoDataFrame: Index with bounding boxes for each row group for each parquet file.
    """
    if not release:
        release = get_newest_release_version()

    _check_release_version(release)
    cache_directory = _get_global_wide_form_release_cache_directory(release)
    index_file_path: Union[Path, str] = cache_directory / _get_wide_form_release_index_file_name(
        theme, type
    )

    file_exists = Path(index_file_path).exists()
    filesystem = None

    if not file_exists:
        if remote_index:
            filesystem = HTTPFileSystem()
            local_cache_path = _get_local_wide_form_release_cache_directory(release)
            index_file_path = LFS_DIRECTORY_URL + str(
                local_cache_path / _get_wide_form_release_index_file_name(theme, type)
            )
        elif skip_index_download:
            # Generate the index and skip download
            generate_wide_form_all_column_names_release_index(
                release,
                verbosity_mode=verbosity_mode,
            )
        else:
            # Try to download the index or generate it if cannot be downloaded
            download_existing_wide_form_all_column_names_release_index(
                release,
                verbosity_mode=verbosity_mode,
            ) or generate_wide_form_all_column_names_release_index(
                release,
                verbosity_mode=verbosity_mode,
            )

    return pd.read_parquet(index_file_path, filesystem=filesystem)


@overload
def download_existing_wide_form_all_column_names_release_index(
    *,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> bool: ...


@overload
def download_existing_wide_form_all_column_names_release_index(
    release: str, *, verbosity_mode: VERBOSITY_MODE = "transient"
) -> bool: ...


def download_existing_wide_form_all_column_names_release_index(
    release: Optional[str] = None,
    *,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> bool:
    """
    Download a pregenerated wide form all column names index for an Overture Maps dataset release.

    Args:
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".

    Returns:
        bool: Information whether index have been downloaded or not.
    """
    return _download_existing_wide_form_all_column_names_release_index(
        release=release, verbosity_mode=verbosity_mode
    )


def _download_existing_wide_form_all_column_names_release_index(
    release: Optional[str] = None,
    theme: Optional[str] = None,
    type: Optional[str] = None,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> bool:
    """
    Download a pregenerated index for an Overture Maps dataset release.

    Args:
        theme (Optional[str], optional): Specify a theme to be downloaded. Defaults to None.
        type (Optional[str], optional): Specify a type to be downloaded. Defaults to None.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".

    Returns:
        bool: Information whether index have been downloaded or not.
    """
    if (theme is None and type is not None) or (theme is not None and type is None):
        raise ValueError("Theme and type both have to be present or None.")

    if not release:
        release = get_newest_release_version()

    _check_release_version(release)

    global_cache_directory = _get_global_wide_form_release_cache_directory(release)
    local_cache_directory = _get_local_wide_form_release_cache_directory(release)

    logger = get_pooch_logger()
    logger.setLevel("WARNING")

    try:
        index_content_file_name = "release_index_content.json"
        index_content_file_url = (
            LFS_DIRECTORY_URL + (local_cache_directory / index_content_file_name).as_posix()
        )
        retrieve(
            index_content_file_url,
            fname=index_content_file_name,
            path=global_cache_directory,
            progressbar=False,
            known_hash=None,
        )

        theme_type_tuples = json.loads(
            (global_cache_directory / index_content_file_name).read_text()
        )

        with TrackProgressBar(verbosity_mode=verbosity_mode) as progress:
            for theme_type_tuple in progress.track(
                theme_type_tuples,
                description="Downloading release indexes",
            ):
                theme_value = theme_type_tuple["theme"]
                type_value = theme_type_tuple["type"]
                sha_value = theme_type_tuple["sha"]
                file_name = _get_wide_form_release_index_file_name(theme_value, type_value)

                if theme_value == theme and type_value == type:
                    continue

                index_file_url = LFS_DIRECTORY_URL + (local_cache_directory / file_name).as_posix()
                retrieve(
                    index_file_url,
                    fname=file_name,
                    path=global_cache_directory,
                    progressbar=False,
                    known_hash=sha_value,
                )

    except HTTPError as ex:
        if ex.response.status_code == 404:
            return False

        raise

    return True


def generate_wide_form_all_column_names_release_index(
    release: str,
    ignore_cache: bool = False,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> bool:
    """
    Generate a wide form all column names index for an Overture Maps dataset release.

    Args:
        release (str): Release version.
        ignore_cache (bool, optional): Whether to ignore precalculated parquet files or not.
            Defaults to False.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".

    Returns:
        bool: Information whether index have been generated or not.
    """
    return _generate_wide_form_all_column_names_release_index(
        release=release, ignore_cache=ignore_cache, verbosity_mode=verbosity_mode
    )


def _generate_wide_form_all_column_names_release_index(
    release: str,
    index_location_path: Optional[Path] = None,
    ignore_cache: bool = False,
    verbosity_mode: VERBOSITY_MODE = "transient",
) -> bool:
    """
    Generate files with all possible wide form column names for an Overture Maps dataset release.

    Args:
        release (str): Release version.
        dataset_path (str, optional): Specify dataset path.
            Defaults to "overturemaps-us-west-2/release/".
        dataset_fs (Literal["local", "s3"], optional): Which filesystem use in PyArrow operations.
            Defaults to "s3".
        index_location_path (Path, optional): Specify index location path. If not, will generate to
            the global cache location. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated parquet files or not.
            Defaults to False.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".

    Returns:
        bool: Information whether index have been generated or not.
    """
    _check_release_version(release)
    cache_directory = Path(
        index_location_path or _get_global_wide_form_release_cache_directory(release)
    )
    release_index_path = cache_directory / "release_index_content.json"

    if release_index_path.exists() and not ignore_cache:
        rprint("Cache exists. Skipping generation.")
        return False

    cache_directory.mkdir(exist_ok=True, parents=True)

    file_hashes = []

    with TrackProgressBar(verbosity_mode=verbosity_mode) as progress:
        for (theme_value, type_value), definition in progress.track(
            sorted(get_theme_type_classification(release=release).items()),
            description="Saving parquet indexes",
        ):
            file_name = _get_wide_form_release_index_file_name(theme_value, type_value)
            cache_file_path = cache_directory / file_name

            if cache_file_path.exists() and not ignore_cache:
                rprint(f"Cache for {theme_value}/{type_value} exists. Skipping generation.")
                file_hashes.append(
                    dict(theme=theme_value, type=type_value, sha=file_hash(str(cache_file_path)))
                )
                continue

            df = definition.get_all_possible_column_names_function(
                theme=theme_value,
                type=type_value,
                release_version=release,
                hierarchy_columns=definition.hierachy_columns,
            )

            df.to_parquet(cache_file_path)
            file_hashes.append(
                dict(theme=theme_value, type=type_value, sha=file_hash(str(cache_file_path)))
            )
            rprint(f"Saved index file {cache_file_path}")

    pd.DataFrame(file_hashes).to_json(release_index_path, orient="records")

    return True


def _get_wide_form_release_index_file_name(theme_value: str, type_value: str) -> str:
    return f"{theme_value}_{type_value}.parquet"



"""
Functions.

This module contains helper functions to simplify the usage.
"""

from pathlib import Path
from typing import Optional, Union, overload

import geopandas as gpd
from shapely import box
from shapely.geometry.base import BaseGeometry

from overturemaestro._rich_progress import VERBOSITY_MODE
from overturemaestro.data_downloader import (
    PYARROW_FILTER,
    download_data,
    download_data_for_multiple_types,
)
from overturemaestro.geopandas_io import geodataframe_from_parquet

__all__ = [
    "convert_bounding_box_to_parquet",
    "convert_bounding_box_to_parquet_for_multiple_types",
    "convert_bounding_box_to_geodataframe",
    "convert_bounding_box_to_geodataframe_for_multiple_types",
    "convert_geometry_to_parquet",
    "convert_geometry_to_parquet_for_multiple_types",
    "convert_geometry_to_geodataframe",
    "convert_geometry_to_geodataframe_for_multiple_types",
]

# TODO: add debug_memory, debug_times
# TODO: prepare examples


@overload
def convert_geometry_to_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@overload
def convert_geometry_to_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: str,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@overload
def convert_geometry_to_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


def convert_geometry_to_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path:
    """
    Get a GeoParquet file with Overture Maps data within given geometry.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single file as a result.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filter (Optional[PYARROW_FILTER], optional): Filters to apply on a pyarrow dataset.
            Can be pyarrow.compute.Expression or List[Tuple] or List[List[Tuple]]. Defaults to None.
        columns_to_download (Optional[list[str]], optional): List of columns to download.
            Automatically adds geometry column to the list. If None, will download all columns.
            Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        Path: Path to the generated GeoParquet file.

    Examples:
        Get buildings in the center of London.

        >>> import overturemaestro as om
        >>> from shapely import box
        >>> london_bbox = box(-0.120077, 51.498164, -0.090809, 51.508849)
        >>> gpq_path = om.convert_geometry_to_parquet(
        ...     release="2024-08-20.0",
        ...     theme="buildings",
        ...     type="building",
        ...     geometry_filter=london_bbox,
        ... ) # doctest: +IGNORE_RESULT
        >>> gpq_path.as_posix()
        'files/2024-08-20.0/theme=buildings/type=building/7ed11f0d_nofilter.parquet'

        Inspect the content
        >>> import geopandas as gpd
        >>> gdf = gpd.read_parquet(gpq_path)
        >>> len(gdf)
        1863
        >>> list(gdf.columns) # doctest: +IGNORE_RESULT
        ['id', 'geometry', 'bbox', 'version', 'sources', 'subtype', 'class', 'names', 'level',
        'has_parts', 'height', 'is_underground', 'num_floors', 'num_floors_underground',
        'min_height', 'min_floor', 'facade_color', 'facade_material', 'roof_material', 'roof_shape',
        'roof_direction', 'roof_orientation', 'roof_color', 'roof_height', 'theme', 'type']

        Download museums in the same area from places dataset with a filter.

        >>> gpq_path = om.convert_geometry_to_parquet(
        ...     release="2024-08-20.0",
        ...     theme="places",
        ...     type="place",
        ...     geometry_filter=london_bbox,
        ...     pyarrow_filter=[[
        ...         (("categories", "primary"), "=", "museum"),
        ...         ("confidence", ">", 0.95),
        ...     ]],
        ... ) # doctest: +IGNORE_RESULT
        >>> gdf = gpd.read_parquet(gpq_path)
        >>> len(gdf)
        5
        >>> gdf[["id", "names", "confidence"]] # doctest: +IGNORE_RESULT
                                         id                               names  confidence
        0  08f194ad1499c8b1038ff3e213d81456  {'primary': 'Florence Nightinga...    0.982253
        1  08f194ad149044c6037575af3681766f  {'primary': 'Philip Simpson Des...    0.969941
        2  08f194ad32a0d494030fdddc1b405fb1  {'primary': 'Shakespeare's Glob...    0.991993
        3  08f194ad30695784036410e184708927  {'primary': 'Clink Street Londo...    0.965185
        4  08f194ad30690a42034312e00c0254a2  {'primary': 'The Clink Prison M...    0.982253
    """
    return download_data(
        release=release,
        theme=theme,
        type=type,
        geometry_filter=geometry_filter,
        pyarrow_filter=pyarrow_filter,
        columns_to_download=columns_to_download,
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )


@overload
def convert_geometry_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@overload
def convert_geometry_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: str,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@overload
def convert_geometry_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


def convert_geometry_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]:
    """
    Get GeoParquet files with Overture Maps data within given geometry for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a list of files as a result.

    Order of paths is the same as the input theme_type_pairs list.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        columns_to_download (Optional[list[Optional[list[str]]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        list[Path]: List of paths to the generated GeoParquet files.
    """
    return download_data_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=geometry_filter,
        release=release,
        pyarrow_filters=pyarrow_filters,
        columns_to_download=columns_to_download,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )


@overload
def convert_geometry_to_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: str,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame: ...


def convert_geometry_to_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame:
    """
    Get a GeoDataFrame with Overture Maps data within given geometry.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single GeoDataFrame as a result.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filter (Optional[PYARROW_FILTER], optional): Filters to apply on a pyarrow dataset.
            Can be pyarrow.compute.Expression or List[Tuple] or List[List[Tuple]]. Defaults to None.
        columns_to_download (Optional[list[str]], optional): List of columns to download.
            Automatically adds geometry column to the list. If None, will download all columns.
            Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.

    Examples:
        Get buildings in the center of London.

        >>> import overturemaestro as om
        >>> from shapely import box
        >>> london_bbox = box(-0.120077, 51.498164, -0.090809, 51.508849)
        >>> gdf = om.convert_geometry_to_geodataframe(
        ...     release="2024-08-20.0",
        ...     theme="buildings",
        ...     type="building",
        ...     geometry_filter=london_bbox,
        ... ) # doctest: +IGNORE_RESULT
        >>> gdf[['names', 'subtype']].sort_index()
                                                                       names       subtype
        id
        08b194ad14804fff0200fea269f9879c  {'primary': 'Park Plaza London ...          None
        08b194ad14812fff02006b5f7b4749e1                                None          None
        08b194ad14814fff02002e44dac80f43  {'primary': 'The Barn', 'common...  agricultural
        08b194ad14814fff0200c77856a66cd7                                None          None
        08b194ad14814fff0200dbc14b9a6d57                                None          None
        ...                                                              ...           ...
        08b194ad33db2fff02006a3ce00700f9  {'primary': 'citizenM hotel Lon...          None
        08b194ad33db3fff02008b05d22745ed  {'primary': 'Metal Box Factory'...          None
        08b194ad33db4fff0200cb2043a25c3c                                None    commercial
        08b194ad33db4fff0200f2ead15d53ac                                None   residential
        08b194ad33db5fff02005eaafd2ff033  {'primary': 'Cooper & Southwark...    commercial
        <BLANKLINE>
        [1863 rows x 2 columns]

        Download museums in the same area from places dataset with a filter.

        >>> gdf = om.convert_geometry_to_geodataframe(
        ...     release="2024-08-20.0",
        ...     theme="places",
        ...     type="place",
        ...     geometry_filter=london_bbox,
        ...     pyarrow_filter=[[
        ...         (("categories", "primary"), "=", "museum"),
        ...         ("confidence", ">", 0.95),
        ...     ]],
        ... ) # doctest: +IGNORE_RESULT
        >>> gdf[["names", "confidence"]].sort_values(by='confidence', ascending=False)
                                                                       names  confidence
        id
        08f194ad32a0d494030fdddc1b405fb1  {'primary': 'Shakespeare's Glob...    0.991993
        08f194ad1499c8b1038ff3e213d81456  {'primary': 'Florence Nightinga...    0.982253
        08f194ad30690a42034312e00c0254a2  {'primary': 'The Clink Prison M...    0.982253
        08f194ad149044c6037575af3681766f  {'primary': 'Philip Simpson Des...    0.969941
        08f194ad30695784036410e184708927  {'primary': 'Clink Street Londo...    0.965185
    """
    parsed_geoparquet_file = download_data(
        theme=theme,
        type=type,
        geometry_filter=geometry_filter,
        release=release,
        pyarrow_filter=pyarrow_filter,
        columns_to_download=columns_to_download,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )
    return geodataframe_from_parquet(parsed_geoparquet_file)


@overload
def convert_geometry_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]: ...


@overload
def convert_geometry_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: str,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]: ...


@overload
def convert_geometry_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]: ...


def convert_geometry_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]:
    """
    Get GeoDataFrames list with Overture Maps data within given geometry for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a list of GeoDataFrames as a result.

    Order of GeoDataFrames is the same as the input theme_type_pairs list.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        columns_to_download (Optional[list[Optional[list[str]]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        list[gpd.GeoDataFrame]: List of GeoDataFrames with Overture Maps features.
    """
    parsed_geoparquet_files = download_data_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=geometry_filter,
        release=release,
        pyarrow_filters=pyarrow_filters,
        columns_to_download=columns_to_download,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )
    return [
        geodataframe_from_parquet(parsed_geoparquet_file)
        for parsed_geoparquet_file in parsed_geoparquet_files
    ]


@overload
def convert_bounding_box_to_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@overload
def convert_bounding_box_to_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@overload
def convert_bounding_box_to_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


def convert_bounding_box_to_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> Path:
    """
    Get a GeoParquet file with Overture Maps data within given bounding box.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single file as a result.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filter (Optional[PYARROW_FILTER], optional): Filters to apply on a pyarrow dataset.
            Can be pyarrow.compute.Expression or List[Tuple] or List[List[Tuple]]. Defaults to None.
        columns_to_download (Optional[list[str]], optional): List of columns to download.
            Automatically adds geometry column to the list. If None, will download all columns.
            Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        Path: Path to the generated GeoParquet file.

    Examples:
        Get buildings in the center of London.

        >>> import overturemaestro as om
        >>> london_bbox = (-0.120077, 51.498164, -0.090809, 51.508849)
        >>> gpq_path = om.convert_bounding_box_to_parquet(
        ...     release="2024-08-20.0",
        ...     theme="buildings",
        ...     type="building",
        ...     bbox=london_bbox,
        ... ) # doctest: +IGNORE_RESULT
        >>> gpq_path.as_posix()
        'files/2024-08-20.0/theme=buildings/type=building/7ed11f0d_nofilter.parquet'

        Inspect the content
        >>> import geopandas as gpd
        >>> gdf = gpd.read_parquet(gpq_path)
        >>> len(gdf)
        1863
        >>> list(gdf.columns) # doctest: +IGNORE_RESULT
        ['id', 'geometry', 'bbox', 'version', 'sources', 'subtype', 'class', 'names', 'level',
        'has_parts', 'height', 'is_underground', 'num_floors', 'num_floors_underground',
        'min_height', 'min_floor', 'facade_color', 'facade_material', 'roof_material', 'roof_shape',
        'roof_direction', 'roof_orientation', 'roof_color', 'roof_height', 'theme', 'type']

        Download museums in the same area from places dataset with a filter.

        >>> gpq_path = om.convert_bounding_box_to_parquet(
        ...     release="2024-08-20.0",
        ...     theme="places",
        ...     type="place",
        ...     bbox=london_bbox,
        ...     pyarrow_filter=[[
        ...         (("categories", "primary"), "=", "museum"),
        ...         ("confidence", ">", 0.95),
        ...     ]],
        ... ) # doctest: +IGNORE_RESULT
        >>> gdf = gpd.read_parquet(gpq_path)
        >>> len(gdf)
        5
        >>> gdf[["id", "names", "confidence"]] # doctest: +IGNORE_RESULT
                                         id                               names  confidence
        0  08f194ad1499c8b1038ff3e213d81456  {'primary': 'Florence Nightinga...    0.982253
        1  08f194ad149044c6037575af3681766f  {'primary': 'Philip Simpson Des...    0.969941
        2  08f194ad32a0d494030fdddc1b405fb1  {'primary': 'Shakespeare's Glob...    0.991993
        3  08f194ad30695784036410e184708927  {'primary': 'Clink Street Londo...    0.965185
        4  08f194ad30690a42034312e00c0254a2  {'primary': 'The Clink Prison M...    0.982253
    """
    return convert_geometry_to_parquet(
        theme=theme,
        type=type,
        geometry_filter=box(*bbox),
        release=release,
        pyarrow_filter=pyarrow_filter,
        columns_to_download=columns_to_download,
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )


@overload
def convert_bounding_box_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@overload
def convert_bounding_box_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@overload
def convert_bounding_box_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


def convert_bounding_box_to_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]:
    """
    Get GeoParquet files with Overture Maps data within given bounding box for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a list of files as a result.

    Order of paths is the same as the input theme_type_pairs list.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        columns_to_download (Optional[list[Optional[list[str]]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        list[Path]: List of paths to the generated GeoParquet files.
    """
    return convert_geometry_to_parquet_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=box(*bbox),
        release=release,
        pyarrow_filters=pyarrow_filters,
        columns_to_download=columns_to_download,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )


@overload
def convert_bounding_box_to_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame: ...


def convert_bounding_box_to_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame:
    """
    Get a GeoDataFrame with Overture Maps data within given bounding box.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single GeoDataFrame as a result.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filter (Optional[PYARROW_FILTER], optional): Filters to apply on a pyarrow dataset.
            Can be pyarrow.compute.Expression or List[Tuple] or List[List[Tuple]]. Defaults to None.
        columns_to_download (Optional[list[str]], optional): List of columns to download.
            Automatically adds geometry column to the list. If None, will download all columns.
            Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.

    Examples:
        Get buildings in the center of London.

        >>> import overturemaestro as om
        >>> london_bbox = (-0.120077, 51.498164, -0.090809, 51.508849)
        >>> gdf = om.convert_bounding_box_to_geodataframe(
        ...     release="2024-08-20.0",
        ...     theme="buildings",
        ...     type="building",
        ...     bbox=london_bbox,
        ... ) # doctest: +IGNORE_RESULT
        >>> gdf[['names', 'subtype']].sort_index()
                                                                       names       subtype
        id
        08b194ad14804fff0200fea269f9879c  {'primary': 'Park Plaza London ...          None
        08b194ad14812fff02006b5f7b4749e1                                None          None
        08b194ad14814fff02002e44dac80f43  {'primary': 'The Barn', 'common...  agricultural
        08b194ad14814fff0200c77856a66cd7                                None          None
        08b194ad14814fff0200dbc14b9a6d57                                None          None
        ...                                                              ...           ...
        08b194ad33db2fff02006a3ce00700f9  {'primary': 'citizenM hotel Lon...          None
        08b194ad33db3fff02008b05d22745ed  {'primary': 'Metal Box Factory'...          None
        08b194ad33db4fff0200cb2043a25c3c                                None    commercial
        08b194ad33db4fff0200f2ead15d53ac                                None   residential
        08b194ad33db5fff02005eaafd2ff033  {'primary': 'Cooper & Southwark...    commercial
        <BLANKLINE>
        [1863 rows x 2 columns]

        Download museums in the same area from places dataset with a filter.

        >>> gdf = om.convert_bounding_box_to_geodataframe(
        ...     release="2024-08-20.0",
        ...     theme="places",
        ...     type="place",
        ...     bbox=london_bbox,
        ...     pyarrow_filter=[[
        ...         (("categories", "primary"), "=", "museum"),
        ...         ("confidence", ">", 0.95),
        ...     ]],
        ... ) # doctest: +IGNORE_RESULT
        >>> gdf[["names", "confidence"]].sort_values(by='confidence', ascending=False)
                                                                       names  confidence
        id
        08f194ad32a0d494030fdddc1b405fb1  {'primary': 'Shakespeare's Glob...    0.991993
        08f194ad1499c8b1038ff3e213d81456  {'primary': 'Florence Nightinga...    0.982253
        08f194ad30690a42034312e00c0254a2  {'primary': 'The Clink Prison M...    0.982253
        08f194ad149044c6037575af3681766f  {'primary': 'Philip Simpson Des...    0.969941
        08f194ad30695784036410e184708927  {'primary': 'Clink Street Londo...    0.965185
    """
    return convert_geometry_to_geodataframe(
        theme=theme,
        type=type,
        geometry_filter=box(*bbox),
        release=release,
        pyarrow_filter=pyarrow_filter,
        columns_to_download=columns_to_download,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )


@overload
def convert_bounding_box_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]: ...


@overload
def convert_bounding_box_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]: ...


@overload
def convert_bounding_box_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]: ...


def convert_bounding_box_to_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> list[gpd.GeoDataFrame]:
    """
    Get GeoDataFrames list with Overture Maps data within given bounding box for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a list of GeoDataFrames as a result.

    Order of GeoDataFrames is the same as the input theme_type_pairs list.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        columns_to_download (Optional[list[Optional[list[str]]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        list[gpd.GeoDataFrame]: List of GeoDataFrames with Overture Maps features.
    """
    return convert_geometry_to_geodataframe_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=box(*bbox),
        release=release,
        pyarrow_filters=pyarrow_filters,
        columns_to_download=columns_to_download,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )



import multiprocessing
from multiprocessing.managers import SyncManager
from pathlib import Path
from queue import Empty, Queue
from time import sleep, time
from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

from overturemaestro._rich_progress import VERBOSITY_MODE, TrackProgressSpinner

if TYPE_CHECKING:  # pragma: no cover
    from multiprocessing.managers import ValueProxy
    from threading import Lock

    import pyarrow as pa
    import pyarrow.fs as fs
    from rich.progress import Progress

# Using `spawn` method to enable integration with Polars and probably other Rust-based libraries
# https://docs.pola.rs/user-guide/misc/multiprocessing/
ctx: multiprocessing.context.SpawnContext = multiprocessing.get_context("spawn")


class MultiprocessingRuntimeError(RuntimeError): ...


def _job(
    queue: Queue[tuple[str, int]],
    tracker: "ValueProxy[int]",
    tracker_lock: "Lock",
    save_path: Path,
    function: Callable[[str, int, "pa.Table"], "pa.Table"],
    columns: Optional[list[str]],
    filesystem: "fs.FileSystem",
) -> None:  # pragma: no cover
    import hashlib

    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    current_pid = multiprocessing.current_process().pid

    writers = {}
    while not queue.empty():
        try:
            file_name, row_group_index = None, None
            file_name, row_group_index = queue.get_nowait()

            fragment_manual = ds.ParquetFileFormat().make_fragment(
                file=file_name,
                filesystem=filesystem,
                row_groups=[row_group_index],
            )
            row_group_table = fragment_manual.to_table(columns=columns)
            if row_group_table.num_rows == 0:
                with tracker_lock:
                    tracker.value += 1
                continue

            result_table = function(file_name, row_group_index, row_group_table)

            if result_table.num_rows == 0:
                with tracker_lock:
                    tracker.value += 1
                continue

            h = hashlib.new("sha256")
            h.update(result_table.schema.to_string().encode())
            schema_hash = h.hexdigest()

            if schema_hash not in writers:
                filepath = save_path / str(current_pid) / f"{schema_hash}.parquet"
                filepath.parent.mkdir(exist_ok=True, parents=True)
                writers[schema_hash] = pq.ParquetWriter(filepath, result_table.schema)

            writers[schema_hash].write_table(result_table)

            with tracker_lock:
                tracker.value += 1
        except Empty:
            pass
        except Exception as ex:
            if file_name is not None and row_group_index is not None:
                queue.put((file_name, row_group_index))

            msg = (
                f"Error in worker (PID: {current_pid},"
                f" Parquet: {file_name}, Row group: {row_group_index})"
            )
            raise MultiprocessingRuntimeError(msg) from ex

    for writer in writers.values():
        writer.close()


class WorkerProcess(ctx.Process):  # type: ignore[name-defined,misc]
    def __init__(self, *args: Any, **kwargs: Any):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = multiprocessing.Pipe()
        self._exception: Optional[tuple[Exception, str]] = None

    def run(self) -> None:  # pragma: no cover
        try:
            multiprocessing.Process.run(self)
            self._cconn.send(None)
        except Exception as e:
            import traceback

            tb: str = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self) -> Optional[tuple[Exception, str]]:
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


class SingletonContextManager(SyncManager):
    def __new__(cls, ctx: multiprocessing.context.SpawnContext) -> "SingletonContextManager":
        if not hasattr(cls, "instance"):
            cls.instance = ctx.Manager()
        return cast(SingletonContextManager, cls.instance)


def _read_row_group_number(path: str, filesystem: "fs.FileSystem") -> int:
    import pyarrow.parquet as pq

    return int(pq.ParquetFile(path, filesystem=filesystem).num_row_groups)


def map_parquet_dataset(
    dataset_path: Union[str, Path, list[str], list[Path]],
    destination_path: Path,
    function: Callable[[str, int, "pa.Table"], "pa.Table"],
    progress_description: str,
    columns: Optional[list[str]] = None,
    filesystem: Optional["fs.FileSystem"] = None,
    report_progress_as_text: bool = True,
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
) -> None:
    """
    Apply a function over parquet dataset in a multiprocessing environment.

    Will save results in multiple files in a destination path.

    Args:
        dataset_path (Union[str, list[str]]): Path(s) of the parquet dataset.
        destination_path (Path): Path of the destination.
        function (Callable[[str, pa.Table], pa.Table]): Function to apply over a parquet file name
            and a row group table. Will save resulting table in a new parquet file.
        progress_description (str): Progress bar description.
        columns (Optional[list[str]], optional): List of columns to read. Defaults to None.
        filesystem (Optional[fs.FileSystem], optional): Filesystem for the dataset.
            Defaults to None.
        report_progress_as_text (bool, optional): Whether to report task progress every minute.
            Defaults to True.
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
    """
    with TrackProgressSpinner(
        "Preparing multiprocessing environment", verbosity_mode=verbosity_mode
    ):
        from concurrent.futures import ProcessPoolExecutor
        from functools import partial

        import pyarrow.parquet as pq

        from overturemaestro._rich_progress import TrackProgressBar

        manager = SingletonContextManager(ctx=ctx)

        queue: Queue[tuple[str, int]] = manager.Queue()
        tracker: ValueProxy[int] = manager.Value("i", 0)
        tracker_lock: Lock = manager.Lock()

        dataset = pq.ParquetDataset(dataset_path, filesystem=filesystem)

        no_cpus = multiprocessing.cpu_count()

        min_no_workers = 32 if no_cpus >= 8 else 16
        no_scan_workers = min(
            max(min_no_workers, no_cpus + 4), 64
        )  # minimum 16 / 32 workers, but not more than 64

        no_processing_workers = no_cpus

        if max_workers:
            no_scan_workers = min(max_workers, no_scan_workers)
            no_processing_workers = min(max_workers, no_processing_workers)

    total_files = len(dataset.files)

    with (
        TrackProgressBar(verbosity_mode=verbosity_mode) as progress,
        ProcessPoolExecutor(max_workers=min(no_scan_workers, total_files)) as ex,
    ):
        fn = partial(_read_row_group_number, filesystem=dataset.filesystem)
        row_group_numbers = list(
            progress.track(
                ex.map(fn, dataset.files, chunksize=1),
                description="Reading all parquet files row groups",
                total=total_files,
            )
        )

        for pq_file, row_group_number in zip(dataset.files, row_group_numbers):
            for row_group in range(row_group_number):
                queue.put((pq_file, row_group))

    total = queue.qsize()

    destination_path.mkdir(parents=True, exist_ok=True)

    try:
        processes = [
            WorkerProcess(
                target=_job,
                args=(
                    queue,
                    tracker,
                    tracker_lock,
                    destination_path,
                    function,
                    columns,
                    dataset.filesystem,
                ),
            )
            for _ in range(min(no_processing_workers, total))
        ]
        with TrackProgressBar(verbosity_mode=verbosity_mode) as progress_bar:
            progress_bar.add_task(description=progress_description, total=total)
            _run_processes(
                processes=processes,
                queue=queue,
                tracker=tracker,
                tracker_lock=tracker_lock,
                total=total,
                progress_bar=progress_bar,
                report_progress_as_text=report_progress_as_text,
            )
    finally:  # pragma: no cover
        _report_exceptions(processes=processes)


def _run_processes(
    processes: list[WorkerProcess],
    queue: Queue[tuple[str, int]],
    tracker: "ValueProxy[int]",
    tracker_lock: "Lock",
    total: int,
    progress_bar: "Progress",
    report_progress_as_text: bool,
) -> None:
    # Run processes
    for p in processes:
        if queue.empty():
            break
        p.start()

    sleep_time = 0.1
    next_update_time = time()
    while any(process.is_alive() for process in processes):
        if any(p.exception for p in processes):  # pragma: no cover
            break

        with tracker_lock:
            completed = tracker.value

        progress_bar.update(task_id=progress_bar.task_ids[0], completed=completed, refresh=True)

        if report_progress_as_text:
            current_time = time()
            if time() >= next_update_time:
                progress_bar.print_progress()
                next_update_time = current_time + 60

        sleep(sleep_time)
        sleep_time = min(1.0, sleep_time + 0.1)

    progress_bar.update(task_id=progress_bar.task_ids[0], completed=total, refresh=True)


def _report_exceptions(processes: list[WorkerProcess]) -> None:
    # In case of exception
    exceptions = []
    for p in processes:
        if p.is_alive():
            p.terminate()

        if p.exception:
            exceptions.append(p.exception)

    if exceptions:
        # use ExceptionGroup in Python3.11
        _raise_multiple(exceptions)


def _raise_multiple(exceptions: list[tuple[Exception, str]]) -> None:
    if not exceptions:
        return
    try:
        error, traceback = exceptions.pop()
        msg = f"{error}\n\nOriginal {traceback}"
        raise type(error)(msg)
    finally:
        _raise_multiple(exceptions)



"""Tests related to generating release indexes."""

import shutil
from pathlib import Path

import geopandas as gpd
import pyarrow.parquet as pq

from overturemaestro import convert_bounding_box_to_parquet_for_multiple_types
from overturemaestro._geometry_clustering import decompress_ranges
from overturemaestro.release_index import (
    _generate_release_index,
    download_existing_release_index,
    get_available_theme_type_pairs,
)

SELECTED_CITIES_BOUNDING_BOXES = [
    (16.8073393, 51.0426686, 17.1762192, 51.2100604),  # Wrocław, Poland
    (13.088345, 52.3382448, 13.7611609, 52.6755087),  # Berlin, Germany
    (126.7644328, 37.4285424, 127.1837702, 37.7014794),  # Seoul, South Korea
    (31.2200331, 29.7483062, 31.9090054, 30.3209168),  # Cairo, Egypt
    (-123.2249611, 49.1989306, -123.0232419, 49.3161714),  # Vancouver, Canada
]


def test_generate_release_indexes(test_release_version: str) -> None:
    """Test if generated indexes are working properly."""
    working_directory = "files/local_overture_maps_dataset"

    # Load data from existing Overture dataset using existing index
    download_existing_release_index(release=test_release_version)
    selected_pairs = [
        pair
        for pair in get_available_theme_type_pairs(release=test_release_version)
        if pair[0] in ("base", "buildings", "places", "transportation")
    ]
    for bounding_box in SELECTED_CITIES_BOUNDING_BOXES:
        convert_bounding_box_to_parquet_for_multiple_types(
            theme_type_pairs=selected_pairs,
            bbox=bounding_box,
            release=test_release_version,
            working_directory=working_directory,
        )

    # Generate local index
    test_index_path = Path("test_index")
    if test_index_path.exists():
        shutil.rmtree(test_index_path)

    _generate_release_index(
        release=test_release_version,
        dataset_path=working_directory,
        dataset_fs="local",
        index_location_path=test_index_path,
    )

    # Check if all row_groups are mapped and if all rows are covered by the index
    for theme_value, type_value in selected_pairs:
        index_gdf = gpd.read_parquet(f"test_index/{theme_value}_{type_value}.parquet")
        unique_files = index_gdf["filename"].unique()

        for unique_file in unique_files:
            assert Path(unique_file).exists()

        dataset_index = (
            index_gdf.explode("row_indexes_ranges")
            .groupby(["filename", "row_group"])["row_indexes_ranges"]
            .apply(list)
            .reset_index()
        )

        row_groups_to_check = dataset_index[
            ["filename", "row_group", "row_indexes_ranges"]
        ].to_dict(orient="records")

        for row_group_to_check in row_groups_to_check:
            local_row_group = pq.ParquetFile(row_group_to_check["filename"]).read_row_group(
                row_group_to_check["row_group"]
            )
            data_index = local_row_group["id"].to_pandas().index
            decompressed_range = decompress_ranges(row_group_to_check["row_indexes_ranges"])

            assert data_index.difference(decompressed_range).empty



"""Tests for public API functions."""

from contextlib import nullcontext as does_not_raise
from typing import Any, Optional

import pytest
from parametrization import Parametrization as P

from overturemaestro._exceptions import MissingColumnError
from overturemaestro.functions import (
    convert_bounding_box_to_geodataframe,
    convert_bounding_box_to_parquet,
)
from overturemaestro.release_index import get_available_theme_type_pairs
from tests.conftest import TEST_RELEASE_VERSION, bbox


@pytest.mark.parametrize(
    "theme_type_pair",
    get_available_theme_type_pairs(TEST_RELEASE_VERSION),
)  # type: ignore
def test_theme_type_pairs(theme_type_pair: tuple[str, str], test_release_version: str) -> None:
    """Test if all theme type pairs are working."""
    convert_bounding_box_to_parquet(
        theme=theme_type_pair[0],
        type=theme_type_pair[1],
        bbox=bbox(),
        release=test_release_version,
        verbosity_mode="verbose",
        ignore_cache=True,
    )


def test_pyarrow_filtering(test_release_version: str) -> None:
    """Test if pyarrow filtering works properly."""
    gdf = convert_bounding_box_to_geodataframe(
        theme="places",
        type="place",
        release=test_release_version,
        bbox=(-0.120077, 51.498164, -0.090809, 51.508849),
        pyarrow_filter=[
            [
                (("categories", "primary"), "=", "museum"),
                ("confidence", ">", 0.95),
            ]
        ],
        verbosity_mode="verbose",
        ignore_cache=True,
    )
    assert (gdf["confidence"] > 0.95).all()
    assert (gdf["categories"].apply(lambda x: x["primary"] == "museum")).all()


@P.parameters("columns_to_download", "expectation")  # type: ignore
@P.case("No columns", [], does_not_raise())  # type: ignore
@P.case("No columns", None, does_not_raise())  # type: ignore
@P.case("Base columns", ["subtype", "class", "id"], does_not_raise())  # type: ignore
@P.case("Base columns without id", ["subtype", "class"], does_not_raise())  # type: ignore
@P.case(
    "Non-existent columns", ["nonexistent_column"], pytest.raises(MissingColumnError)
)  # type: ignore
def test_columns_download(
    test_release_version: str, columns_to_download: Optional[list[str]], expectation: Any
) -> None:
    """Test if columns download restricting works properly."""
    with expectation:
        gdf = convert_bounding_box_to_geodataframe(
            theme="buildings",
            type="building",
            release=test_release_version,
            bbox=(-0.120077, 51.498164, -0.090809, 51.508849),
            columns_to_download=columns_to_download,
            verbosity_mode="verbose",
            ignore_cache=True,
        )
        assert "geometry" in gdf.columns
        assert all(
            column_name in gdf.columns
            for column_name in (columns_to_download or [])
            if column_name != "id"  # skip indexed column
        )


def test_empty_region(test_release_version: str) -> None:
    """Test if regions without data are properly parsed."""
    gdf = convert_bounding_box_to_geodataframe(
        "places", "place", (-10, -10, -9.9, -9.9), release=test_release_version
    )

    assert gdf.empty
    assert all(
        c in gdf.columns
        for c in (
            "geometry",
            "bbox",
            "version",
            "sources",
            "names",
            "categories",
            "confidence",
            "websites",
            "socials",
            "emails",
            "phones",
            "brand",
            "addresses",
        )
    )
